package filegateway

import (
	"context"
	"encoding/json"
	"testing"
	"text/template"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	"xbus.io/go-xbus/v4/envelope"
	"xbus.io/go-xbus/v4/testutils"
)

func newTestEmit() *testEmit {
	return &testEmit{}
}

type testEmit struct {
	Emitted []envelope.Envelope
}

func (t *testEmit) Emit(
	ctx context.Context, env envelope.Envelope,
) error {
	t.Emitted = append(t.Emitted, env)
	return nil
}

func TestEmitter(t *testing.T) {
	t.Run("basics", func(t *testing.T) {
		var (
			fname1 = t.Name() + "/f1.txt"
			fname2 = t.Name() + "/f2.log"
			fname3 = t.Name() + "/f3.txt"
		)
		fs := getMemoryFilestem(".")
		fs.pollingInterval = 5 * time.Millisecond
		logger := testutils.GetTestLogger(t)
		testEmit := newTestEmit()

		emitter := Emitter{
			emitEnvelope: testEmit.Emit,
			log:          logger.XbusLogger,
			fs:           fs,
			pattern:      t.Name() + "/*.txt",
			msgType:      "test",
			postAction:   "delete",
		}

		require.NoError(t, emitter.Startup())

		time.Sleep(10 * time.Millisecond)
		require.NoError(t, writeFile(fs, fname1, "f1 content"))

		time.Sleep(10 * time.Millisecond)
		require.NoError(t, writeFile(fs, fname2, "f2 content"))

		time.Sleep(10 * time.Millisecond)
		require.NoError(t, writeFile(fs, fname3, "f3 content"))

		time.Sleep(20 * time.Millisecond)

		require.NoError(t, emitter.Shutdown())

		if l := len(testEmit.Emitted); l != 2 {
			t.Fatalf("Expected 2 emissions, got %d", l)
		}

		env1, err := envelope.Complete(testEmit.Emitted[0], time.Second)
		require.NoError(t, err)
		env2, err := envelope.Complete(testEmit.Emitted[1], time.Second)
		require.NoError(t, err)
		var m1, m2 stdlib.FileMeta
		require.NoError(t, json.Unmarshal(env1.Events[0].Items[0], &m1))
		require.NoError(t, json.Unmarshal(env2.Events[0].Items[0], &m2))
		assert.Equal(t, "f1.txt", m1.Filename)
		assert.Equal(t, "f1 content", string(env1.Events[0].Items[1]))
		assert.Equal(t, "f3.txt", m2.Filename)
		assert.Equal(t, "f3 content", string(env2.Events[0].Items[1]))
		assert.NotContains(t, fs.Files, fname1)
		assert.Contains(t, fs.Files, fname2)
		assert.NotContains(t, fs.Files, fname3)
	})

	t.Run("post-action 'rename'", func(t *testing.T) {
		var (
			fname1 = t.Name() + "/f1.txt"
		)
		fs := NewMemoryFilesystem(".")
		fs.pollingInterval = 5 * time.Millisecond
		logger := testutils.GetTestLogger(t)
		testEmit := newTestEmit()

		emitter := Emitter{
			emitEnvelope: testEmit.Emit,
			log:          logger.XbusLogger,
			fs:           fs,
			pattern:      t.Name() + "/*.txt",
			msgType:      "test",
			postAction:   "rename",
			renameTemplate: template.Must(template.New("rename").Parse(
				"{{ .Dir }}/sub/{{ .WithoutExt }}-{{ .EnvelopeID }}-post{{ .Ext }}")),
		}

		require.NoError(t, emitter.Startup())

		require.NoError(t, writeFile(fs, fname1, "f1 content"))

		time.Sleep(20 * time.Millisecond)

		require.NoError(t, emitter.Shutdown())

		if l := len(testEmit.Emitted); l != 1 {
			t.Errorf("Expected 1 emissions, got %d", l)
			if l < 1 {
				t.FailNow()
			}
		}

		env1, err := envelope.Complete(testEmit.Emitted[0], time.Second)
		require.NoError(t, err)
		var m1 stdlib.FileMeta
		require.NoError(t, json.Unmarshal(env1.Events[0].Items[0], &m1))

		assert.NotContains(t, fs.Files, fname1)
		assert.Equal(t, 1, len(fs.Files))

		for name := range fs.Files {
			assert.Regexp(t,
				t.Name()+"/sub/f1-([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}){1}-post.txt", name)
		}
	})

	t.Run("emptyFile", func(t *testing.T) {
		var (
			fname = t.Name() + "/f1.txt"
		)

		fs := getMemoryFilestem(".")
		fs.pollingInterval = 5 * time.Millisecond
		logger := testutils.GetTestLogger(t)
		testEmit := newTestEmit()

		emitter := Emitter{
			emitEnvelope: testEmit.Emit,
			log:          logger.XbusLogger,
			fs:           fs,
			pattern:      t.Name() + "/*.txt",
			msgType:      "test",
			postAction:   "delete",
		}

		require.NoError(t, emitter.Startup())

		require.NoError(t, writeFile(fs, fname, ""))

		time.Sleep(20 * time.Millisecond)

		require.NoError(t, emitter.Shutdown())

		require.Equal(t, 1, len(testEmit.Emitted))

		env, err := envelope.Complete(testEmit.Emitted[0], time.Second)
		require.NoError(t, err)
		var meta stdlib.FileMeta
		require.NoError(t, json.Unmarshal(env.Events[0].Items[0], &meta))
		assert.Equal(t, "f1.txt", meta.Filename)
		if len(env.Events[0].Items) == 2 {
			assert.Equal(t, "", string(env.Events[0].Items[1]))
		} else {
			assert.Equal(t, 1, len(env.Events[0].Items))
		}
	})

	t.Run("bigFile", func(t *testing.T) {
		var (
			fname       = t.Name() + "/f1.txt"
			contentSize = 1024 * 1024 * 50
			fcontent    = make([]byte, contentSize)
		)
		for i := range fcontent {
			fcontent[i] = byte(48 + i%10)
		}

		fs := getMemoryFilestem(".")
		fs.pollingInterval = 5 * time.Millisecond
		logger := testutils.GetTestLogger(t)
		testEmit := newTestEmit()

		emitter := Emitter{
			emitEnvelope: testEmit.Emit,
			log:          logger.XbusLogger,
			fs:           fs,
			pattern:      t.Name() + "/*.txt",
			msgType:      "test",
			postAction:   "delete",
		}

		require.NoError(t, emitter.Startup())

		require.NoError(t, writeFileBytes(fs, fname, fcontent))

		time.Sleep(20 * time.Millisecond)

		require.NoError(t, emitter.Shutdown())

		require.Equal(t, 1, len(testEmit.Emitted))

		env, err := envelope.Complete(testEmit.Emitted[0], time.Second)
		require.NoError(t, err)
		var meta stdlib.FileMeta
		require.NoError(t, json.Unmarshal(env.Events[0].Items[0], &meta))
		assert.Equal(t, "f1.txt", meta.Filename)

		var count int
		for _, item := range env.Events[0].Items[1:] {
			assert.Equal(t, (fcontent[count : count+len(item)]), (item))
			count += len(item)
		}
	})
}
