package stdlib

import (
	"bytes"
	"io"
	"log"
	"os"
	"testing"

	minio "github.com/minio/minio-go"
	"github.com/ory/dockertest"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func pingMinio(endpoint, accessKey, secretKey string) error {
	minioClient, err := minio.New(
		endpoint, accessKey, secretKey, false)
	if err != nil {
		return err
	}
	_, err = minioClient.ListBuckets()
	return err
}

func TestFileLocationS3(t *testing.T) {
	accessKey := "Iaqu4pai0hoh"
	secretKey := "thaeShoon0we"
	bucketname := "testbucket"
	objectid := "testobject"
	endpoint := ""

	inCI := os.Getenv("CI") == "true"
	if inCI {
		accessKey = os.Getenv("MINIO_ACCESS_KEY")
		secretKey = os.Getenv("MINIO_SECRET_KEY")
		endpoint = os.Getenv("MINIO_PORT_9000_TCP_ADDR") +
			":" + os.Getenv("MINIO_PORT_9000_TCP_PORT")
	} else {
		pool, err := dockertest.NewPool("")
		if err != nil {
			log.Fatalf("Could not connect to docker: %s", err)
		}

		// pulls an image, creates a container based on it and runs it
		resource, err := pool.RunWithOptions(
			&dockertest.RunOptions{
				Repository: "minio/minio",
				Tag:        "RELEASE.2019-06-19T18-24-42Z",
				Env: []string{
					"MINIO_ACCESS_KEY=" + accessKey,
					"MINIO_SECRET_KEY=" + secretKey,
				},
				Cmd: []string{"server", "/data"},
			})
		if err != nil {
			log.Fatalf("Could not start resource: %s", err)
		}
		defer func() { assert.NoError(t, pool.Purge(resource)) }()
		minioport := resource.GetPort("9000/tcp")
		endpoint = "localhost:" + minioport

		require.NoError(t, pool.Retry(func() error {
			return pingMinio(endpoint, accessKey, secretKey)
		}))
	}

	minioClient, err := minio.New(
		endpoint, accessKey, secretKey, false)
	require.NoError(t, err)

	require.NoError(t, minioClient.MakeBucket(bucketname, ""))

	fileContent := bytes.NewBuffer([]byte("Hello World!"))
	if _, err := minioClient.PutObject(
		bucketname, objectid,
		fileContent, int64(fileContent.Len()),
		minio.PutObjectOptions{},
	); err != nil {
		t.Fatal(err)
	}

	SetS3Endpoint(endpoint, S3Settings{
		AccessKeyID:     accessKey,
		SecretAccessKey: secretKey,
		UseSSL:          false,
	})

	t.Run("ReadBody", func(t *testing.T) {
		location := FileLocation{Type: "s3",
			Args: map[string]interface{}{
				"provider":    endpoint,
				"bucket-name": bucketname,
				"objectid":    objectid,
			}}

		locationType := GetLocationType("s3")
		reader, err := locationType.ReadBody(location)
		if err != nil {
			t.Fatal(err)
		}
		defer reader.Close()
		data, err := io.ReadAll(reader)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, "Hello World!", string(data))
	})
}
