package routing

import (
	"context"

	libjq_go "github.com/flant/libjq-go"
	"github.com/flant/libjq-go/pkg/jq"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	"orus.io/xbus/xbus-stdlib/service"

	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// NewUpdater initialize a Updater
func NewUpdater(actor *api.Actor, settings service.Settings) *Updater {
	return &Updater{
		settings: settings,
	}
}

// NewUpdaterAsWorker initialize a Updater and returns it as ActorService
func NewUpdaterAsWorker(actor *api.Actor, settings service.Settings, log xbus.Logger) (service.Worker, error) {
	return NewUpdater(actor, settings), nil
}

// Updater is a Worker that updates the routing table
type Updater struct {
	settings service.Settings

	extractKeyJq   *jq.JqProgram
	extractValueJq *jq.JqProgram

	table TableStore
}

// Startup setup the routing table and compiles the jq programs
func (u *Updater) Startup() error {
	table, err := NewTableStore("store", u.settings)
	if err != nil {
		return err
	}
	extractKeyJq, err := u.settings.String("extract-key-jq")
	if err != nil {
		return err
	}
	extractKeyJqProg, err := libjq_go.Jq().Program(extractKeyJq).Precompile()
	if err != nil {
		return err
	}
	extractValueJq, err := u.settings.String("extract-value-jq")
	if err != nil {
		return err
	}
	extractValueJqProg, err := libjq_go.Jq().Program(extractValueJq).Precompile()
	if err != nil {
		return err
	}
	u.extractKeyJq = extractKeyJqProg
	u.extractValueJq = extractValueJqProg
	u.table = table
	return nil
}

// Shutdown unsuscribes and close the DB connection
func (u *Updater) Shutdown() error {
	return u.table.Close()
}

// Process ...
func (u *Updater) Process(
	ctx context.Context, req *api.ActorProcessRequest,
	inputs service.Inputs, outputs service.Outputs, log xbus.Logger,
) error {
	env, err := inputs.Get(api.DefaultInput)
	if err != nil {
		return err
	}

	_, body, err := stdlib.UniqueStringMessageContext(ctx, env)
	if err != nil {
		return err
	}

	key, err := u.extractKeyJq.RunRaw(body)
	if err != nil {
		return err
	}

	value, err := u.extractValueJq.RunRaw(body)
	if err != nil {
		return err
	}

	if err := u.table.Set(key, value); err != nil {
		return err
	}

	return outputs.Forward(api.DefaultOutput, env.ID())
}

func init() {
	service.MustRegisterWorker("routing.table-updater", NewUpdaterAsWorker)
}
